
from anthill.common import admin as a
from anthill.common import to_int

from . model.content import ContentError, ContentNotFound
from . model.promo import PromoError, PromoNotFound

import ujson
import datetime


class RootAdminController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.links("Promo service", [
                a.link("contents", "Edit contents", icon="paper-plane"),
                a.link("promos", "Edit promo codes", icon="gift")
            ])
        ]


class ContentsController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([], "Contents"),
            a.links("Items", [
                a.link("content", item.name, icon="paper-plane", content_id=item.content_id)
                for item in data["items"]
            ]),
            a.links("Navigate", [
                a.link("index", "Go back", icon="chevron-left"),
                a.link("new_content", "Create content", icon="plus")
            ])
        ]

    async def get(self):
        contents = self.application.contents
        items = await contents.list_contents(self.gamespace)

        result = {
            "items": items
        }

        return result


class ContentController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([
                a.link("contents", "Contents")
            ], "Content"),
            a.form("Update content", fields={
                "content_name": a.field("Content unique ID", "text", "primary", "non-empty"),
                "content_json": a.field("Content payload (any useful data)", "json", "primary", "non-empty")
            }, methods={
                "update": a.method("Update", "primary", order=1),
                "delete": a.method("Delete this content", "danger", order=2)
            }, data=data),
            a.links("Navigate", [
                a.link("contents", "Go back", icon="chevron-left")
            ])
        ]

    async def get(self, content_id):

        contents = self.application.contents

        try:
            content = await contents.get_content(self.gamespace, content_id)
        except ContentNotFound:
            raise a.ActionError("No such content")

        result = {
            "content_name": content.name,
            "content_json": content.payload
        }

        return result

    async def update(self, content_name, content_json):

        content_id = self.context.get("content_id")

        try:
            content_json = ujson.loads(content_json)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted JSON")

        contents = self.application.contents

        try:
            await contents.update_content(self.gamespace, content_id, content_name, content_json)
        except ContentError as e:
            raise a.ActionError("Failed to update content: " + e.args[0])

        raise a.Redirect(
            "content",
            message="Content has been updated",
            content_id=content_id)

    # noinspection PyUnusedLocal
    async def delete(self, **ignored):

        content_id = self.context.get("content_id")
        contents = self.application.contents

        try:
            await contents.delete_content(self.gamespace, content_id)
        except ContentError as e:
            raise a.ActionError("Failed to delete content: " + e.args[0])

        raise a.Redirect("contents", message="Content has been deleted")


class NewContentController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([
                a.link("contents", "Contents")
            ], "New contents"),
            a.form("New content", fields={
                "content_name": a.field("Content unique ID", "text", "primary", "non-empty"),
                "content_json": a.field("Content payload (any useful data)", "json", "primary", "non-empty")
            }, methods={
                "create": a.method("Create", "primary")
            }, data={"content_json": {}}),
            a.links("Navigate", [
                a.link("contents", "Go back", icon="chevron-left")
            ])
        ]

    async def create(self, content_name, content_json):

        try:
            content_json = ujson.loads(content_json)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted JSON")

        contents = self.application.contents

        try:
            content_id = await contents.new_content(self.gamespace, content_name, content_json)
        except ContentError as e:
            raise a.ActionError("Failed to create new content: " + e.args[0])

        raise a.Redirect(
            "content",
            message="New content has been created",
            content_id=content_id)


class PromosController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([], "Promo codes"),
            a.form(title="Edit promo code", fields={
                "code": a.field("Edit promo code", "text", "primary", "non-empty")
            }, methods={
                "edit": a.method("Edit", "primary")
            }, data=data),
            a.links("Navigate", [
                a.link("index", "Go back", icon="chevron-left"),
                a.link("new_promo", "Create a new promo code", icon="plus"),
                a.link("new_promos", "Create multiple promo codes", icon="plus-square")
            ])
        ]

    async def edit(self, code):
        promos = self.application.promos

        try:
            promos.validate(code)
        except PromoError as e:
            raise a.ActionError(e.message)

        try:
            promo = await promos.find_promo(self.gamespace, code)
        except PromoNotFound:
            raise a.ActionError("No such promo code")

        raise a.Redirect("promo", promo_id=promo.code_id)


class NewPromoController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([
                a.link("promos", "Promo codes")
            ], "New promo code"),
            a.form("New promo code", fields={
                "promo_key": a.field("Promo code key", "text", "primary", "non-empty"),
                "promo_amount": a.field("Promo uses amount", "text", "primary", "number"),
                "promo_expires": a.field("Expire date", "date", "primary", "non-empty"),
                "promo_contents": a.field("Promo items", "kv", "primary", "non-empty",
                                          values=data["content_items"])
            }, methods={
                "create": a.method("Create", "primary")
            }, data=data),
            a.links("Navigate", [
                a.link("contents", "Go back", icon="chevron-left")
            ])
        ]

    async def get(self):

        contents = self.application.contents
        content_items = {
            item.content_id: item.name
            for item in (await contents.list_contents(self.gamespace))
        }

        return {
            "promo_key": "<random>",
            "promo_amount": "1",
            "content_items": content_items,
            "promo_expires": str(datetime.datetime.now() + datetime.timedelta(days=30))
        }

    async def create(self, promo_key, promo_amount, promo_expires, promo_contents):
        promos = self.application.promos

        try:
            promo_contents = ujson.loads(promo_contents)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted JSON")

        if promo_key == "<random>":
            promo_key = promos.random()

        try:
            promos.validate(promo_key)
        except PromoError as e:
            raise a.ActionError(e.message)

        try:
            promo_id = await promos.new_promo(self.gamespace, promo_key, promo_amount, promo_expires, promo_contents)
        except ContentError as e:
            raise a.ActionError("Failed to create new promo: " + e.args[0])

        raise a.Redirect(
            "promo",
            message="Promo code has been created",
            promo_id=promo_id)


class NewPromosController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        res = [
            a.breadcrumbs([
                a.link("promos", "Promo codes")
            ], "New promo codes")
        ]

        result = data.get("result", None)

        if result:
            res.append(a.form("Generated keys", fields={
                "result": a.field("Keys", "text", "primary", "non-empty", multiline=20)
            }, methods={}, data=data))
        else:
            res.append(a.form("New promo code", fields={
                "promo_keys": a.field("Number of keys to generate", "text", "primary", "number"),
                "promo_amount": a.field("Promo uses amount", "text", "primary", "number"),
                "promo_expires": a.field("Expire date", "date", "primary", "non-empty"),
                "promo_contents": a.field("Promo items", "kv", "primary", "non-empty",
                                          values=data["content_items"])
            }, methods={
                "create": a.method("Create", "primary")
            }, data=data))

        res.append(a.links("Navigate", [
                a.link("contents", "Go back", icon="chevron-left")
            ]))

        return res

    async def get(self):

        contents = self.application.contents
        content_items = {
            item.content_id: item.name
            for item in (await contents.list_contents(self.gamespace))
        }

        return {
            "promo_keys": "2",
            "promo_amount": "1",
            "content_items": content_items,
            "promo_expires": str(datetime.datetime.now() + datetime.timedelta(days=30))
        }

    async def create(self, promo_keys, promo_amount, promo_expires, promo_contents):
        promos = self.application.promos

        promo_keys = to_int(promo_keys)

        try:
            promo_contents = ujson.loads(promo_contents)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted JSON")

        result = []

        for i in range(0, promo_keys):
            promo_key = promos.random()

            try:
                promos.validate(promo_key)
            except PromoError as e:
                raise a.ActionError(e.message)

            try:
                await promos.new_promo(self.gamespace, promo_key, promo_amount, promo_expires, promo_contents)
            except ContentError:
                continue
            else:
                result.append(promo_key)

        return {
            "result": "\n".join(result)
        }


class PromoController(a.AdminController):
    def access_scopes(self):
        return ["promo_admin"]

    def render(self, data):
        return [
            a.breadcrumbs([
                a.link("promos", "Promo codes")
            ], "Promo code '{0}'".format(data["promo_code"])),
            a.form("Update promo code", fields={
                "promo_code": a.field("Promo code key", "text", "primary", "non-empty"),
                "promo_amount": a.field("Usage amount left", "text", "primary", "number"),
                "promo_expires": a.field("Expire date", "date", "primary", "non-empty"),
                "promo_contents": a.field("Promo items", "kv", "primary", "non-empty",
                                          values=data["content_items"])
            }, methods={
                "update": a.method("Update", "primary"),
                "delete": a.method("Delete this promo code", "danger")
            }, data=data),
            a.links("Accounts used this promo code", [a.link(
                "/profile/profile", "@" + account, account=account) for account in data["usages"]]),
            a.links("Navigate", [
                a.link("contents", "Go back", icon="chevron-left")
            ])
        ]

    async def get(self, promo_id):

        promos = self.application.promos
        contents = self.application.contents
        content_items = {
            item.content_id: item.name
            for item in (await contents.list_contents(self.gamespace))
        }

        try:
            promo = await promos.get_promo(self.gamespace, promo_id)
        except PromoNotFound:
            raise a.ActionError("No such promo code")

        try:
            usages = await promos.get_promo_usages(self.gamespace, promo_id)
        except PromoError as e:
            raise a.ActionError(e.message)

        result = {
            "promo_code": promo.key,
            "promo_amount": promo.amount,
            "promo_contents": promo.contents,
            "content_items": content_items,
            "promo_expires": str(promo.expires),
            "usages": usages
        }

        return result

    async def update(self, promo_code, promo_amount, promo_expires, promo_contents):

        promo_id = self.context.get("promo_id")

        promos = self.application.promos

        try:
            promo_contents = ujson.loads(promo_contents)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted JSON")

        try:
            await promos.update_promo(self.gamespace, promo_id, promo_code, promo_amount, promo_expires,
                                      promo_contents)
        except ContentError as e:
            raise a.ActionError("Failed to update promo code: " + e.args[0])

        raise a.Redirect("promo", message="Promo code has been updated", promo_id=promo_id)

    # noinspection PyUnusedLocal
    async def delete(self, **ignored):

        promo_id = self.context.get("promo_id")
        promos = self.application.promos

        try:
            await promos.delete_promo(self.gamespace, promo_id)
        except ContentError as e:
            raise a.ActionError("Failed to delete promo: " + e.args[0])

        raise a.Redirect("promos", message="Promo code has been deleted")




from anthill.common.database import DatabaseError, DuplicateError
from anthill.common.model import Model

import ujson
import re
import random


class PromoError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return str(self.code) + ": " + self.message


class PromoNotFound(Exception):
    pass


class PromoExists(Exception):
    pass


class PromoAdapter(object):
    def __init__(self, data):
        self.code_id = str(data.get("code_id"))
        self.key = data.get("code_key")
        self.expires = data.get("code_expires")
        self.contents = data.get("code_contents")
        self.amount = data.get("code_amount")


class PromoModel(Model):
    PROMO_PATTERN = re.compile("[A-Z0-9]{4}-[A-Z0-9]{4}-[A-Z0-9]{4}")

    def __init__(self, db):
        self.db = db

    def get_setup_db(self):
        return self.db

    def get_setup_tables(self):
        return ["promo_code", "promo_code_users"]

    def random_code(self, n):
        return ''.join(random.choice("ABCDEFGHJKLMNPQRSTUVWXYZ0123456789") for _ in range(n))

    def random(self):
        return self.random_code(4) + "-" + self.random_code(4) + "-" + self.random_code(4)

    def validate(self, code):
        if not re.match(PromoModel.PROMO_PATTERN, code):
            raise PromoError(400, "Promo code is not valid (should be XXXX-XXXX-XXXX)")

    def has_delete_account_event(self):
        return True

    async def accounts_deleted(self, gamespace, accounts, gamespace_only):
        try:
            if gamespace_only:
                await self.db.execute(
                    """
                        DELETE FROM `promo_code_users`
                        WHERE `gamespace_id`=%s AND `account_id` IN %s;
                    """, gamespace, accounts)
            else:
                await self.db.execute(
                    """
                        DELETE FROM `promo_code_users`
                        WHERE `account_id` IN %s;
                    """, accounts)
        except DatabaseError as e:
            raise PromoError(500, "Failed to delete promo code usages: " + e.args[1])

    async def wrap_contents(self, gamespace_id, contents):
        keys = list(contents.keys())

        try:
            wrapped = await self.db.query("""
                SELECT * FROM `promo_contents`
                WHERE `gamespace_id`=%s AND  `content_name` IN %s;
            """, gamespace_id, keys)
        except DatabaseError as e:
            raise PromoError(500, "Failed to wrap contents: " + e.args[1])

        result = {
            str(item["content_id"]): contents[item["content_name"]]
            for item in wrapped
        }

        return result

    async def new_promo(self, gamespace_id, promo_key, promo_use_amount, promo_expires, promo_contents):

        if not isinstance(promo_contents, dict):
            raise PromoError(400, "Contents is not a dict")

        try:
            await self.find_promo(gamespace_id, promo_key)
        except PromoNotFound:
            pass
        else:
            raise PromoError(409, "Promo code '{0}' already exists.".format(promo_key))

        try:
            result = await self.db.insert("""
                INSERT INTO `promo_code`
                (`gamespace_id`, `code_key`, `code_amount`, `code_expires`, `code_contents`)
                VALUES (%s, %s, %s, %s, %s);
            """, gamespace_id, promo_key, promo_use_amount, promo_expires, ujson.dumps(promo_contents))
        except DuplicateError:
            raise PromoExists()
        except DatabaseError as e:
            raise PromoError(500, "Failed to add new promo code: " + e.args[1])

        return result

    async def find_promo(self, gamespace_id, promo_key):
        try:
            result = await self.db.get("""
                SELECT *
                FROM `promo_code`
                WHERE `code_key`=%s AND `gamespace_id`=%s;
            """, promo_key, gamespace_id)
        except DatabaseError as e:
            raise PromoError(500, "Failed to find promo code: " + e.args[1])

        if result is None:
            raise PromoNotFound()

        return PromoAdapter(result)

    async def get_promo(self, gamespace_id, promo_id):
        try:
            result = await self.db.get("""
                SELECT *
                FROM `promo_code`
                WHERE `code_id`=%s AND `gamespace_id`=%s;
            """, promo_id, gamespace_id)
        except DatabaseError as e:
            raise PromoError(500, "Failed to get promo code: " + e.args[1])

        if result is None:
            raise PromoNotFound()

        return PromoAdapter(result)

    async def delete_promo(self, gamespace_id, promo_id):
        try:
            await self.db.execute("""
                DELETE
                FROM `promo_code`
                WHERE `code_id`=%s AND `gamespace_id`=%s;
            """, promo_id, gamespace_id)

            await self.db.execute("""
                DELETE
                FROM `promo_code_users`
                WHERE `code_id`=%s AND `gamespace_id`=%s;
            """, promo_id, gamespace_id)
        except DatabaseError as e:
            raise PromoError(500, "Failed to delete content: " + e.args[1])

    async def update_promo(self, gamespace_id, promo_id, promo_key, promo_use_amount, promo_expires, promo_contents):

        if not isinstance(promo_contents, dict):
            raise PromoError(400, "Contents is not a dict")

        try:
            await self.db.execute("""
                UPDATE `promo_code`
                SET `code_key`=%s, `code_amount`=%s, `code_expires`=%s, `code_contents`=%s
                WHERE `code_id`=%s AND `gamespace_id`=%s;
            """, promo_key, promo_use_amount, promo_expires, ujson.dumps(promo_contents), promo_id, gamespace_id)
        except DatabaseError as e:
            raise PromoError(500, "Failed to update content: " + e.args[1])

    async def get_promo_usages(self, gamespace_id, promo_id):
        usages = await self.db.query("""
            SELECT `account_id`
            FROM `promo_code_users`
            WHERE `code_id`=%s AND `gamespace_id`=%s;
        """, promo_id, gamespace_id)

        return [str(usage["account_id"]) for usage in usages]

    async def use_promo(self, gamespace_id, account_id, promo_key):
        async with self.db.acquire(auto_commit=False) as db:
            try:
                promo = await db.get(
                    """
                        SELECT `code_id`, `code_contents`, `code_amount`
                        FROM `promo_code`
                        WHERE `code_key`=%s AND `gamespace_id`=%s AND `code_amount` > 0 AND `code_expires` > NOW()
                        FOR UPDATE;
                    """, promo_key, gamespace_id)

                if not promo:
                    raise PromoNotFound()

                promo_id = promo["code_id"]
                promo_contents = promo["code_contents"]
                promo_amount = promo["code_amount"]

                ids = list(promo_contents.keys())

                if not ids:
                    raise PromoError(400, "Promo code has no contents.")

                used = await db.get(
                    """
                        SELECT *
                        FROM `promo_code_users`
                        WHERE `code_id`=%s AND `gamespace_id`=%s AND `account_id`=%s;
                    """, promo_id, gamespace_id, account_id)

                if used:
                    raise PromoError(409, "Code already used by this user")

                await db.insert(
                    """
                        INSERT INTO `promo_code_users`
                        (`gamespace_id`, `code_id`, `account_id`)
                        VALUES (%s, %s, %s);
                    """, gamespace_id, promo_id, account_id)

                promo_amount -= 1

                await db.execute(
                    """
                        UPDATE `promo_code`
                        SET `code_amount` = %s
                        WHERE `code_id`=%s AND `gamespace_id`=%s;
                    """, promo_amount, promo_id, gamespace_id)

                contents_result = []

                contents = await db.query(
                    """
                        SELECT `content_json`, `content_id`
                        FROM `promo_contents`
                        WHERE `content_id` IN %s
                    """, ids)

                for cnt in contents:
                    result = {
                        "payload": cnt["content_json"],
                        "amount": promo_contents[str(cnt["content_id"])]
                    }
                    contents_result.append(result)

                return {
                    "result": contents_result
                }

            finally:
                await db.commit()




from anthill.common.options import options
from anthill.common import server, database, access

from . import handlers as h
# noinspection PyUnresolvedReferences
from . import options as _opts
from . import admin

from . model.content import ContentModel
from . model.promo import PromoModel


class PromoServer(server.Server):
    def __init__(self):
        super(PromoServer, self).__init__()

        self.db = database.Database(
            host=options.db_host,
            database=options.db_name,
            user=options.db_username,
            password=options.db_password)

        self.contents = ContentModel(self.db)
        self.promos = PromoModel(self.db)

    def get_models(self):
        return [self.contents, self.promos]

    def get_handlers(self):
        return [
            (r"/use/(.*)", h.UsePromoHandler),
        ]

    def get_internal_handler(self):
        return h.InternalHandler(self)

    def get_admin(self):
        return {
            "index": admin.RootAdminController,
            "contents": admin.ContentsController,
            "content": admin.ContentController,
            "new_content": admin.NewContentController,
            "promos": admin.PromosController,
            "new_promo": admin.NewPromoController,
            "new_promos": admin.NewPromosController,
            "promo": admin.PromoController
        }

    def get_metadata(self):
        return {
            "title": "Promo",
            "description": "Reward users with promo-codes",
            "icon": "gift"
        }


if __name__ == "__main__":
    stt = server.init()
    access.AccessToken.init([access.public()])
    server.start(PromoServer)




from tornado.web import HTTPError

from anthill.common.access import scoped, AccessToken
from anthill.common.handler import AuthenticatedHandler
from anthill.common.validate import validate
from anthill.common.internal import InternalError

from . model.promo import PromoNotFound, PromoError, PromoExists


class UsePromoHandler(AuthenticatedHandler):
    @scoped(scopes=["promo"])
    async def post(self, promo_key):
        promos = self.application.promos
        gamespace_id = self.token.get(AccessToken.GAMESPACE)

        try:
            promo_usage = await promos.use_promo(gamespace_id, self.token.account, promo_key)
        except PromoError as e:
            raise HTTPError(e.code, e.message)
        except PromoNotFound as e:
            raise HTTPError(404, str(e))
        else:
            self.dumps(promo_usage)


class InternalHandler(object):
    def __init__(self, application):
        self.application = application

    @validate(gamespace="int", amount="int", codes_count="int", expires="datetime", contents="json_dict")
    async def generate_code(self, gamespace, amount, expires, contents, codes_count=1):

        promos = self.application.promos

        contents = await promos.wrap_contents(gamespace, contents)

        keys = []

        for i in range(0, codes_count):
            while True:
                promo_key = promos.random()

                try:
                    await promos.new_promo(gamespace, promo_key, amount, expires, contents)
                except PromoExists:
                    continue
                except PromoError as e:
                    raise InternalError(e.code, e.message)
                else:
                    keys.append(promo_key)
                    break

        return {
            "keys": keys
        }

    @validate(gamespace="int", account="int", key="str")
    async def use_code(self, gamespace, account, key):
        promos = self.application.promos

        try:
            promo_usage = await promos.use_promo(gamespace, account, key)
        except PromoError as e:
            raise InternalError(e.code, e.message)
        except PromoNotFound as e:
            raise InternalError(404, str(e))
        else:
            return promo_usage

    @validate(gamespace="int")
    async def list_contents(self, gamespace):
        contents = self.application.contents

        try:
            items = await contents.list_contents(gamespace)
        except PromoError as e:
            raise InternalError(e.code, e.message)
        except PromoNotFound as e:
            raise InternalError(404, str(e))
        else:
            return {
                "items": {
                    item.content_id: item.name
                    for item in items
                }
            }

    @validate(gamespace="int", promo_key="str")
    async def get_code_info(self, gamespace, promo_key):
        promos = self.application.promos

        try:
            promo = await promos.find_promo(gamespace, promo_key)
        except PromoError as e:
            raise InternalError(e.code, e.message)
        except PromoNotFound as e:
            raise InternalError(404, "No such promo code")
        else:
            return {
                "code": {
                    "expires": str(promo.expires),
                    "id": promo.code_id,
                    "contents": promo.contents,
                    "amount": promo.amount,
                }
            }

    @validate(gamespace="int", code_id="int")
    async def list_code_users(self, gamespace, code_id):
        promos = self.application.promos

        try:
            users = await promos.get_promo_usages(gamespace, code_id)
        except PromoError as e:
            raise InternalError(e.code, e.message)
        else:
            return {
                "users": users
            }




from anthill.common.database import DatabaseError, DuplicateError
from anthill.common.model import Model
import ujson


class ContentError(Exception):
    pass


class ContentNotFound(Exception):
    pass


class ContentAdapter(object):
    def __init__(self, data):
        self.content_id = str(data.get("content_id"))
        self.name = data.get("content_name")
        self.payload = data.get("content_json")


class ContentModel(Model):
    def __init__(self, db):
        self.db = db

    def get_setup_tables(self):
        return ["promo_contents"]

    def get_setup_db(self):
        return self.db

    async def new_content(self, gamespace_id, content_name, content_data):

        try:
            result = await self.db.insert("""
                INSERT INTO `promo_contents`
                (`gamespace_id`, `content_name`, `content_json`)
                VALUES (%s, %s, %s);
            """, gamespace_id, content_name, ujson.dumps(content_data))
        except DuplicateError:
            raise ContentError("Content '{0}' already exists.".format(content_name))
        except DatabaseError as e:
            raise ContentError("Failed to add new content: " + e.args[1])

        return result

    async def find_content(self, gamespace_id, content_name):
        try:
            result = await self.db.get("""
                SELECT *
                FROM `promo_contents`
                WHERE `content_name`=%s AND `gamespace_id`=%s
                LIMIT 1;
            """, content_name, gamespace_id)
        except DatabaseError as e:
            raise ContentError("Failed to find content: " + e.args[1])

        if result is None:
            raise ContentNotFound()

        return ContentAdapter(result)

    async def get_content(self, gamespace_id, content_id):
        try:
            result = await self.db.get("""
                SELECT *
                FROM `promo_contents`
                WHERE `content_id`=%s AND `gamespace_id`=%s
                LIMIT 1;
            """, content_id, gamespace_id)
        except DatabaseError as e:
            raise ContentError("Failed to get content: " + e.args[1])

        if result is None:
            raise ContentNotFound()

        return ContentAdapter(result)

    async def delete_content(self, gamespace_id, content_id):
        try:
            await self.db.execute("""
                DELETE
                FROM `promo_contents`
                WHERE `content_id`=%s AND `gamespace_id`=%s
                LIMIT 1;
            """, content_id, gamespace_id)
        except DatabaseError as e:
            raise ContentError("Failed to delete content: " + e.args[1])

    async def update_content(self, gamespace_id, content_id, content_name, content_data):
        try:
            await self.db.execute("""
                UPDATE `promo_contents`
                SET `content_name`=%s, `content_json`=%s
                WHERE `content_id`=%s AND `gamespace_id`=%s;
            """, content_name, ujson.dumps(content_data), content_id, gamespace_id)
        except DatabaseError as e:
            raise ContentError("Failed to update content: " + e.args[1])

    async def list_contents(self, gamespace_id):
        try:
            contents = await self.db.query("""
                SELECT *
                FROM `promo_contents`
                WHERE `gamespace_id`=%s;
            """, gamespace_id)
        except DatabaseError as e:
            raise ContentError("Failed to list content: " + e.args[1])

        return list(map(ContentAdapter, contents))
